
package URI::URL;

use Carp;
require Cwd;

require Exporter;
@ISA = qw(Exporter);
@EXPORT_OK = qw(uri_escape uri_unescape);

# Define default unsafe characters.
# Note that you cannot reliably change this at runtime
# because the substitutions which use it use the /o flag.
# XXX Should we include '~' or leave it to applications to
# add that if required?.
my $DefaultUnsafe = '\x00-\x20"#%;<>?\x7F-\xFF';

# Basic lexical elements, taken from RFC1738:
# (these are refered to by comments in the code)
# safe         = "$" | "-" | "_" | "." | "+"
# extra        = "!" | "*" | "'" | "(" | ")" | ","
# national     = "{" | "}" | "|" | "\" | "^" | "~" | "[" | "]" | "`"
# punctuation  = "<" | ">" | "#" | "%" | <">
# reserved     = ";" | "/" | "?" | ":" | "@" | "&" | "="
# escape       = "%" hex hex
# unreserved   = alpha | digit | safe | extra
# uchar        = unreserved | escape
# xchar        = unreserved | reserved | escape


$Debug = 0;             # set to 1 to print URL's on creation
my $UseCache = 1;       # see as_string method
my $StrictSchemes = 1;        # see new()

# schemes we have initialised:
my %ImplementedBy = ( '_generic' => 'URI::URL::_generic' );
# clases we have initialised:
my $Implementors  = ();

# Build a hex<->char map (HexHex->Char and Char->HexHex)
map
{
    my($hex, $chr) = (sprintf("%%%02X", $_), chr($_));
    $escapes{$hex} = $chr; $escapes{lc($hex)} = $chr;
    $escapes{$chr} = $hex;
} 0..255;

use strict qw(subs refs);

sub new
{
    my($class, $init, $base) = @_;

    my $this;
    if (ref $init)
    {
        $this = $init->clone;
        $this->base($base) if $base;
     }
     else
     {
         $init =~ s/^\s+//;  # remove leading space
         $init =~ s/\s.*//;  # remove anything after first word
         # We need a scheme to determine which class to use
         my($scheme) = $init =~ m/^([.+\-\w]+):/;
         if (!$scheme and $base)
	 { # get scheme from base
             if (ref $base)
	     { # may be object or just a string
                 $scheme = $base->scheme;
             }
	     else
	     {
                 $scheme = $1 if $base =~ m/^([.+\-\w]+):/;
             }
         }
         unless($scheme)
	 {
             croak "Unable to determine scheme for '$init'"
               if $StrictSchemes;
             $scheme = 'http';
         }
         my $impclass = URI::URL->implementor($scheme);
         unless ($impclass)
	 {
             croak "URI::URL scheme '$scheme' is not supported"
                 if $StrictSchemes;
             $impclass = URI::URL->implementor; # use generic
         }

         # hand-off to scheme specific implementation sub-class
         $this = $impclass->new($init, $base);
     }
     return $this;
}

sub clone
{
    my $self = shift;
    bless { %$self }, ref $self;
}

sub newlocal
{
    my($class, $path) = @_;
    my $url = new URI::URL "file://localhost/";

    unless (defined $path and $path =~ m:^/:)
    {
        my $cwd = Cwd::fastcwd();
        $cwd =~ s:/?$:/:; # force trailing slash on dir
        $path = (defined $path) ? $cwd . $path : $cwd;
    }
    $url->path($path);
    $url;
}

sub print_on
{
    my $self = shift;
    my $fh = shift || 'STDOUT';
    my($k, $v);
    print $fh "Dump of $self...\n";
    foreach $k (sort keys %$self)
    {
        $v = $self->{$k};
        $v = 'UNDEF' unless defined $v;
        print $fh "  $k\t'$v'\n";
    }
}

sub implementor
{
    my($class, $scheme, $impclass) = @_;
    my $ic;
    $scheme = (defined $scheme) ? lc($scheme) : '_generic';

    if ($impclass)
    {
        $impclass->_init_implementor($scheme);
        $ImplementedBy{$scheme} = $impclass;
    }
    return $ic if $ic=$ImplementedBy{$scheme};

    # scheme not yet known, look for internal implementation
    $ic = "URI::URL::$scheme";        # default location
    no strict qw(refs);
    # check we actually have one for the scheme:
    $ic = '' unless defined @{"${ic}::ISA"};
    $ImplementedBy{$scheme} = $ic if $ic;
    $ic->_init_implementor;
    $ic;
}

sub _init_implementor                 # private method
{
    my($class) = @_;
    # Remember that one implementor class may actually
    # serve to implement several URL schemes.

    # have we already initialised this class?
    return 1 if exists $Implementors{$class};

    no strict 'refs';
    # Setup overloading - experimental
    %{"${class}::OVERLOAD"} = %URI::URL::_generic::OVERLOAD
       unless defined %{"${class}::OVERLOAD"};
    $Implementors{$class} = 1;
    1;
}

#####################################################################
#
# Methods to handle URL's elements

# These methods always return the current value,
# so you can use $url->scheme to read the current value.
# If a new value is passed, e.g. $url->scheme('http'),
# it also sets the new value, and returns the previous value.
# Use $url->scheme(undef) to set the value to undefined.

# Generic-RL components:
sub scheme      { shift->elem('scheme', @_);  }
sub netloc      { shift->elem('netloc', @_);  }
sub path        { shift->elem('path',   @_);  }
sub params      { shift->elem('params', @_);  }
sub query       { shift->elem('query',  @_);  }
sub frag        { shift->elem('frag',   @_);  }

# Fields derived from generic netloc:
sub user        { shift->elem('user',   @_);  }
sub password    { shift->elem('password',@_); }
sub host        { shift->elem('host',   @_);  }
sub port        { shift->elem('port',   @_);  }

# optimisation to speed up elem() below:
my %netloc_fields = qw(user 1 password 1 host 1 port 1);

sub elem
{
    my($self, $element, @val) = @_;
    my $old = $self->{$element};
    return $old unless @val;

    $self->{$element} = $val[0]; # general case
    $self->{'_str'} = '';        # void cached string

    # netloc includes user, password, host and port
    if ($element eq 'netloc')
    {
        $self->_read_netloc();  # update parts from whole
    }
    elsif (exists $netloc_fields{$element})
    {
        $self->_write_netloc(); # update whole from parts
    }
    return $old;
}

# Other attributes of a URL object:
# (These may happen to use elem() but applications should not
# use elem() to access them).

sub base
{
    my $self = shift;
    return $self->elem('_base', @_) if @_;      # set
    # The base attribute supports 'lazy' conversion from URL strings
    # to URL objects. Strings may be stored but when a string is
    # fetched it will automatically be converted to a URL object.
    # The main benefit is to make it much cheaper to say:
    #   new URI::URL $random_url_string 'http:'
    my $base = $self->elem('_base');            # get
    return undef unless defined $base;
    unless (ref $base)
    {
        $base = new URI::URL $base;
        $self->elem('_base', $base); # set new object
    }
    $base;
}

sub unsafe
{
    shift->elem('_unsafe', @_);
}

#####################################################################
#
# escape()
# unescape()
#
#  Generic escaping ('this has spaces' -> 'this%20has%20spaces')
#    and unescaping ('this%20has%20spaces' -> 'this has spaces')
#  Overridden by subclasses which need more control.
#
sub escape
{
    my($self, $text, $patn) = @_;
    if ($patn)
    {
        $text =~ s/([$patn])/$URI::URL::escapes{$1}/eg;
        return $text;
    }
    # let perl pre-compile this default for max speed
    $text =~ s/([$DefaultUnsafe])/
        $URI::URL::escapes{$1}/oeg;
    $text;
}

# Define method aliases so that subclasses can control escaping at
# a finer granularity. Doing it this way has practically zero cost.
# Only of significant value to classes which rely on the default
# full_path() and as_string() methods.
*_esc_netloc = \&escape;
*_esc_path   = \&escape;
*_esc_params = \&escape;
*_esc_frag   = \&escape;
sub _esc_query
{
    my($self, $text, @unsafe) = @_;
    $text =~ s/ /+/g; # RFC1630
    my $text = $self->escape($text, @unsafe);
}


sub unescape
{
    my($self, $text) = @_;
    return undef unless defined $text;
    # Note from RFC1630:  "Sequences which start with a percent sign
    # but are not followed by two hexadecimal characters are reserved
    # for future extension"
    $text =~ s/%([\dA-Fa-f][\dA-Fa-f])/chr(hex($1))/eg;
    $text;
}


#####################################################################
#
# Miscellaneous functions (NON-METHODS)

# uri_escape()
#
# Apply URI character escaping rules to some text.
# Note that it is generally better to do something like this:
#      $url = new URI::URL 'http:';
#       $url->path($random_query);
# See the 'HOW AND WHEN TO ESCAPE' section in the pod text above.
#
sub uri_escape
{
    URI::URL->escape(@_);
}

# uri_unescape()
#
# Unescape some text destined to be a component of a URL.
# Note that it is generally better to do something like this:
#       $url->path(uri_unescape($pre_escaped_path));
# See the 'HOW AND WHEN TO ESCAPE' section in the pod text above.
#
sub uri_unescape
{
    URI::URL->unescape(@_);
}


#####################################################################
#
#       Internal pre-defined generic scheme support
#
# In this implementation all schemes are subclassed from
# URI::URL::_generic. This turns out to have reasonable mileage.
# See also draft-ietf-uri-relative-url-05.txt

package URI::URL::_generic;            # base support for generic-RL's

use Carp;
@ISA = qw(URI::URL);
%OVERLOAD = ( '""'=>'as_string', 'fallback'=>1 );     # EXPERIMENTAL

sub new                                 # inherited by subclasses
{      
    my($class, $init, $base) = @_;
    my $url = bless {}, $class;         # create empty object
    $url->_parse($init);                # parse $init into components
    $url->base($base) if $base;
    $url->print_on('STDERR') if $URI::URL::Debug;
    $url;
}


sub _parse
{
    my($self, $u) = @_;
    $self->{'_orig_url'} = $u if $URI::URL::Debug;
    $self->{'frag'}   = $self->unescape($1) if $u =~ s/#(.*)$//;
    $self->{'scheme'} = lc($1)   if $u =~ s/^\s*([\w\+\.\-]+)://;
    $self->{'netloc'} = $self->unescape($1) if $u =~ s!^//([^/]*)!!;
    if ($u =~ s/\?(.*)//)     # '+' -> ' ' for queries (RFC1630)
    {
        my $query = $1;
        $query =~ s/\+/ /g;
        $self->{'query'}  = $self->unescape($query)
    }
    $self->{'params'} = $self->unescape($1) if $u =~ s/;(.*)//;
    $self->{'path'}   = $self->unescape($u);
    # read netloc components: "<user>:<password>@<host>:<port>"
    $self->_read_netloc;
    1;
}

sub _read_netloc       # netloc -> user, password, host, post
{
    my($self) = @_;
    my $nl = $self->{'netloc'} || ''; # already unescaped
    $self->{'_str'} = '';       # void cache
    if ($nl =~ s/^([^:@]*):?(.*?)@//)
    {
        $self->{'user'}     = $1;
        $self->{'password'} = $2 if $2 ne '';
    }
    if ($nl =~ s/^([^:]*):?(\d*)//)
    {
        $self->{'host'} = $1;
        $self->{'port'} = $2 if $2 ne '';
    }
}


sub _write_netloc      # user, password, host, post -> netloc
{
    my($self) = @_;
    my $tmp;
    my $nl = $self->{'user'} || '';
    $nl .= ":$self->{'password'}" if $nl and $self->{'password'};
    $nl .= "\@" if $nl;
    $nl .= ($tmp = $self->{'host'});
    $nl .= ":$tmp" if ($tmp && ($tmp=$self->{'port'})
                            && $tmp != $self->default_port);
    $self->{'netloc'} = $nl;
}


# Generic-RL stringify
#
sub as_string
{
    my $self = shift;
    return $self->{'_str'} if $self->{'_str'} and
        $UseCache;

    # use @ here to avoid undef warnings and allow $self->escape
    # to use optimised pattern if no override has been set.
    my @unsafe = shift || $self->unsafe || ();
    my($scheme, $netloc, $port) = @{$self}{qw(scheme netloc port)};

    # full_path() -> path+query+params+frag (escaped)
    my $path  = $self->full_path(@unsafe);

    if ($netloc)
    {
        $path = "/$path" unless $path =~ m:^/:;
        $path = "//".$self->_esc_netloc($netloc, @unsafe).$path;
    }
    my $urlstr = ($scheme) ? "$scheme:$path" : $path;
    $self->{'_str'} = $urlstr;  # set cache
    return $urlstr;
}


# Generic-RL stringify full path (path+query+params+frag)
#
sub full_path
{
    my $self = shift;
    # use @ here to avoid undef warnings and allow $self->escape
    # to use optimised pattern if no override has been set.
    my @unsafe = shift || $self->unsafe || ();
    my($path, $params, $query, $frag)
        = @{$self}{qw(path params query frag) };
    my $u = '';
    $u .=     $self->_esc_path($path,    @unsafe) if $path;
    $u .= ";".$self->_esc_params($params,@unsafe) if $params;
    $u .= "?".$self->_esc_query($query,  @unsafe) if $query;
    $u .= "#".$self->_esc_frag($frag,    @unsafe) if $frag;
    return $u;
}


# Generic-RL: Resolving Relative URL into an Absolute URL
#
#
sub abs
{
    my($self, $base) = @_;
    my $embed = $self->clone;

    $base = $self->base unless $base;      # default to default base
    return $embed unless $base;            # we have no base 

    $base = new URI::URL $base unless ref $base; # make obj if needed

    my @u = @{$embed}{qw(scheme host port path params query frag) };

    # just use base if we are empty  
    # XXX can we ever be empty? I think scheme is always defined.
    return $base->clone if (scalar(grep($_, @u)) == 0);

    my($scheme, $host, $port, $path, $params, $query, $frag) = @u;

    # if we have a scheme we must already be absolute  
    return $embed if $scheme;

    $embed->{'scheme'} = $base->{'scheme'};  

    return $embed if $embed->{'netloc'};  
    $embed->{'netloc'} = $base->{'netloc'}; 
    $embed->_read_netloc();

    return $embed if $embed->{'path'} =~ m:^/:; 
    
    unless ($embed->{'path'})
    {         
        $embed->{'path'} = $base->{'path'};   

        return $embed if $embed->params;   
        $embed->{'params'} = $base->{'params'}; 

        return $embed if $embed->query;   
        $embed->{'query'} = $base->{'query'};   
        return $embed;
    }

    my $basepath = $base->{'path'};
    my $relpath  = $embed->{'path'};

    $basepath =~ s!^/!!;
    $basepath =~ s!/$!/.!;              # prevent empty segment
    my @path = split('/', $basepath);   # base path into segments
    pop(@path);                         # remove last segment

    $relpath =~ s!/$!/.!;               # prevent empty segment

    push(@path, split('/', $relpath));  # append relative segments

    my @newpath = ();
    my $isdir = 0;
    foreach $segment (@path)
    { 
 #      warn '> ', join('/', @newpath), ": $segment\n";
        if ($segment eq '.')    # ignore "same" directory
	{
            $isdir = 1;
        }
        elsif ($segment eq '..')
	{
            $isdir = 1;
            my $last = pop(@newpath);
            if (!defined $last)   # nothing to pop
	    {
                push(@newpath, $segment); # so must append
            }
            elsif ($last eq '..')   # '..' cannot match '..'
	    {
                # so put back again, and append
                push(@newpath, $last, $segment);
            }
            else
	    {
                # it was a component,
                # keep popped
            }
        }
	else
	{
            $isdir = 0;
            push(@newpath, $segment);
        }
    }

    $embed->{'path'} = join('/', @newpath) . ($isdir ? '/' : '');
    
    return $embed;
}


# default_port()
#
# subclasses will usually want to override this
#
sub default_port
{
    0;
}


# _expect()
#
# Handy low-level object method tester. See test code at end.
#
sub _expect
{
    my($self, $method, $expect, @args) = @_;
    my $result = $self->$method(@args);
    $expect = 'UNDEF' unless defined $expect;
    $result = 'UNDEF' unless defined $result;

    return 1 if $expect eq $result;
    warn "'$self'->$method(@args) = '$result' " .
                "(expected '$expect')\n";
    $self->print_on('STDERR');
    confess "Test Failed";
}

####################################################################

package URI::URL::file;         @ISA = qw(URI::URL::_generic);

# fileurl        = "file://" [ host | "localhost" ] "/" fpath
# fpath          = fsegment *[ "/" fsegment ]
# fsegment       = *[ uchar | "?" | ":" | "@" | "&" | "=" ]
# Note that fsegment can contain '?' (query) but not ';' (param)

sub _parse
{
    my($self, $init) = @_;
    # allow the generic parser to do the bulk of the work
    $self->URI::URL::_generic::_parse($init);
    # then just deal with the effect of rare stray '?'s
    if (defined $self->{'query'})
    {
        $self->{'path'} .= "?$self->{'query'}";
        delete $self->{'query'};
    }
    1;
}

sub _esc_path
{
    my($self, $text) = @_;
    $text =~ s/([^-a-zA-Z\d\$_.+!*'(),%?:@&=\/])/
        $URI::URL::escapes{$1}/oeg;
    $text;
}



package URI::URL::ftp;          @ISA = qw(URI::URL::_generic);

sub default_port { 21 };



package URI::URL::telnet;       @ISA = qw(URI::URL::_generic);

sub default_port { 23 };



package URI::URL::whois;        @ISA = qw(URI::URL::_generic);

sub default_port { 43 };



package URI::URL::gopher;       @ISA = qw(URI::URL::_generic);

sub default_port { 70 };
sub gtype  { shift->elem('gtype', @_); }

sub _parse
{
    my($self, $url) = @_;
    $self->{'scheme'} = lc($1) if $url =~ s/^\s*([\w\+\.\-]+)://;
    $self->{'netloc'} = $self->unescape($1)
                                if $url =~ s!^//([^/]*)!!;
    $self->{'gtype'}  = $self->unescape($1) if $url =~ s!^/(.)!!;
    my @sss = split(/%09/, $url, 3);
    $self->{'selector'} = $self->unescape(shift @sss);
    $self->{'search'}   = $self->unescape(shift @sss);
    $self->{'string'}   = $self->unescape(shift @sss);
}



package URI::URL::finger;       @ISA = qw(URI::URL::_generic);

sub default_port { 79 };



package URI::URL::http;         @ISA = qw(URI::URL::_generic);

sub default_port { 80 };



package URI::URL::nntp;         @ISA = qw(URI::URL::_generic);

sub default_port { 119 };

sub _parse
{
    my($self, $init) = @_;
    $self->URI::URL::_generic::_parse($init);
    my @parts = split(/\//, $self->{'path'});
    $self->{'group'} = $self->unescape($parts[1]);
    $self->{'digits'}= $self->unescape($parts[2]);
}



package URI::URL::news;         @ISA = qw(URI::URL::_generic);

sub _parse
{
    my($self, $init) = @_;
    $self->{'scheme'}  = lc($1) if ($init =~ s/^\s*([\w\+\.\-]+)://);
    my $tmp = $self->unescape($init);
    $self->{'grouppart'} = $tmp;
    $self->{ ($tmp =~ m/\@/) ? 'article' : 'group' } = $tmp;
}



package URI::URL::wais;         @ISA = qw(URI::URL::_generic);

sub default_port { 210 };

sub _parse
{
    my($self, $init) = @_;
    $self->URI::URL::_generic::_parse($init);
    my @parts = split(/\//, $self->{'path'});
    $self->{'database'} = $self->unescape($parts[1]);
    $self->{'wtype'}    = $self->unescape($parts[2]);
    $self->{'wpath'}    = $self->unescape($parts[3]);
}



package URI::URL::webster;      @ISA = qw(URI::URL::_generic);

sub default_port { 765 };



package URI::URL::prospero;     @ISA = qw(URI::URL::_generic);

sub default_port { 1525 };      # says rfc1738, section 3.11



package URI::URL::mailto;       @ISA = qw(URI::URL::_generic);

sub _parse
{
    my($self, $init) = @_;
    $self->{'scheme'}  = lc($1) if ($init =~ s/^\s*([\w\+\.\-]+)://);
    $self->{'encoded822addr'} = $self->unescape($init);
}



package URI::URL::rlogin;       @ISA = qw(URI::URL::_generic);



package URI::URL::tn3270;       @ISA = qw(URI::URL::_generic);



# Aliases for old method names. To be deleted in a future version.
package URI::URL::_generic;

no strict 'refs';
*{"dump"} = \&print_on;
*{"str"}  = \&as_string;


1;

